/* port_driver.c */

#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "erl_driver.h"
#include "ei.h"

#define DRV_LISTEN 1

typedef struct {
    ErlDrvPort port;
} example_data;

static ErlDrvTermData am_ok;
static ErlDrvTermData am_error;

static FILE *debug_file;

#define INIT_ATOM(NAME) am_ ## NAME = (ErlDrvTermData)driver_mk_atom((char*)#NAME)

static void
example_drv_ok(example_data *drv, ErlDrvTermData pid)
{
    ErlDrvTermData spec[2];
	
	spec[0] = ERL_DRV_ATOM;
	spec[1] = am_ok;
    driver_send_term(drv->port, pid, spec, sizeof(spec)/sizeof(spec[0]));
}

static ErlDrvData example_drv_start(ErlDrvPort port, char *buff)
{
	example_data* d;

	debug_file = fopen("logfile.log", "a+");
    d = (example_data*)driver_alloc(sizeof(example_data));
	fprintf(debug_file, "[tcp_drv] example_drv_start\n");fflush(NULL);
    d->port = port;
	set_port_control_flags(port, PORT_CONTROL_FLAG_BINARY);
    return (ErlDrvData)d;
}

static void example_drv_stop(ErlDrvData handle)
{
    driver_free((char*)handle);
}

static int tcp_listen(int port)
{
	struct sockaddr_in servaddr;

	int main_socket = socket(AF_INET, SOCK_STREAM, 0);

	memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family      = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port        = htons(port);

	if ( bind(main_socket, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0 ) {
		fprintf(stderr, "ECHOSERV: Error calling bind()\n");fflush(NULL);
    }

	if ( listen(main_socket, 5) < 0 ) {
		fprintf(stderr, "ECHOSERV: Error calling listen()\n");fflush(NULL);
    }

	fprintf(debug_file, "[tcp_drv] accepting connections on port 5555\n");fflush(NULL);

	return main_socket;
}

static void
example_drv_ready_input(ErlDrvData drv_data, ErlDrvEvent event)
{
    /*example_data *d = (example_data*)drv_data;*/

	fprintf(debug_file, "[tcp_drv] ready input on socket %d\n", (int)event);fflush(NULL);
}

static void
example_drv_outputv(ErlDrvData drv_data, ErlIOVec *ev)
{
    example_data *d = (example_data*)drv_data;
    ErlDrvBinary* data = ev->binv[1];
	/*char* bytes = data->orig_bytes;*/
    unsigned char cmd  = data->orig_bytes[0];
	int listening_socket;

    switch (cmd) {
		case DRV_LISTEN:
			listening_socket = tcp_listen(5555);
			break;
	}

	driver_select(d->port, (ErlDrvEvent) listening_socket, ERL_DRV_READ, 1);

	fprintf(debug_file, "[tcp_drv] waiting for new connections on socket %d\n", listening_socket);fflush(NULL);

	example_drv_ok(d, driver_caller(d->port));
}

int example_drv_driver_init(void)
{
    INIT_ATOM(ok);
    INIT_ATOM(error);

    return 0;
}

static void 
example_drv_process_exit(ErlDrvData drv_data, ErlDrvMonitor* monitor)
{
}

static ErlDrvEntry example_driver_entry = {
    example_drv_driver_init,            /* init */
    example_drv_start,                  /* startup (defined below) */
    example_drv_stop,                   /* shutdown (defined below) */
    NULL,                               /* output */
    example_drv_ready_input,            /* ready_input */
    NULL,								/* ready_output */
    (char*)"tcp_drv",                   /* driver name */
    NULL,                               /* finish */
    NULL,                               /* handle */
    NULL,								/* control */
    NULL,                               /* timeout */
    example_drv_outputv,                /* outputv, binary output */
    NULL,								/* ready_async */
    NULL,                               /* flush */
    NULL,                               /* call */
    NULL,								/* event */
    ERL_DRV_EXTENDED_MARKER,            /* ERL_DRV_EXTENDED_MARKER */
    ERL_DRV_EXTENDED_MAJOR_VERSION,     /* ERL_DRV_EXTENDED_MAJOR_VERSION */
    ERL_DRV_EXTENDED_MINOR_VERSION,     /* ERL_DRV_EXTENDED_MINOR_VERSION */
    ERL_DRV_FLAG_USE_PORT_LOCKING,      /* ERL_DRV_FLAGs */
    NULL,                               /* handle2 (reserved */
    example_drv_process_exit,           /* process_exit */
    NULL                                /* stop_select */
};

DRIVER_INIT(example_drv) /* must match name in driver_entry */
{
    return &example_driver_entry;
}
